<html>
    <head>
        <title>${self.title()}</title>
        ${self.style()}
        ${self.javascript()}
    </head>
    <body>
        ${self.body()}
    </body>
</html>

<%def name="title()">
    Default Title
</%def>

<%def name="style()">
    <link rel="stylesheet" type="text/css" media="screen" href="/css/style.css" />
</%def>

<%def name="javascript()">
    <script language="text/javascript" src="/javascript/shared.js"></script>
</%def>
